(*

                       Driver8 Realtime 3D Rendering Engine
            Copyright (c) 1997, Alex Chalfin. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*)

{
  The clipping process used here is based on a 90 degree field of view so
  simplified bounds equations can be used for clipping tests.

  Upon entry, all verticies are in the 90 degree fustrum no matter what
  field of view is defined by the programmer (the CamZFac in the view fixes
  this). The view fustrum is also assumed to be square (nxn) whereas the
  screen is nxm (320x200 for example). The square correction is also performed
  by the code (using CamYFac).

  All "unique" features are performed by earlier parts of the code and are
  undone in the perspective transform.

  For more information on the actual process, see Computer Graphics:
  Principals and Practics
}
Unit Clip;
{$I DEFINE.INC}

Interface

Uses
  Types;

Procedure ClipPolygons(Var View : ViewObject; Var Obj : VectorObject);

Implementation

Uses
  Shade,
  XForm,
  Pool;

{ Types for n-gon used in clipping }
Type
  CVert3d = Array[0..15] of pVertex;
  CVert2d = Array[0..15] of fVertex2d;
  pCVert3d = ^CVert3d;
  pCVert2d = ^CVert2d;

(*
 * Internal arrays specific to the clipping process.
 *)
Var
  IVerts : CVert3d;
  OVerts : CVert3d;
  IUV    : CVert2d;
  OUV    : CVert2d;

(*
 * Internal functions specific to the clipping process.
 *)

Procedure FustrumClip(View : ViewObject; P : pTriangle; ShadeIndex, Mask : Longint); Forward;
Procedure BottomClip(InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint); Forward;
Procedure RightClip (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint); Forward;
Procedure LeftClip  (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint); Forward;
Procedure TopClip   (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint); Forward;
Procedure zClip     (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint); Forward;

(*
 * Main Clipping function.
 *)
Procedure ClipPolygons(Var View : ViewObject; Var Obj : VectorObject);

Var
  Flag1, Flag2, Flag3 : Boolean;
  ClipFlag   : Longint;
  TotalClip  : Longint;
  i          : Longint;
  P          : pTriangle;
  Av, Bv, Cv : pVertex3d;

Begin
  { Check to see if the object neds to be clipped at all }
  if ((Obj.ClipMask And Not(CLIP_SHADEMASK)) = 0)
    Then Exit;

  For i := 0 to Obj.NumPoly - 1 do
    Begin
      P := @Obj.Polygons^[i];

      { first check of the polygon is visible (no need to clip }
      { non-visible polygons)                                  }
      if (P^.Visible = 1)
        Then Begin
          ClipFlag  := Obj.ClipMask And CLIP_SHADEMASK;
          TotalClip := 0;

          Av := @P^.A^.Cam;
          Bv := @P^.B^.Cam;
          Cv := @P^.C^.Cam;

          if ((Obj.ClipMask And CLIP_Z) <> 0)
            Then Begin
              Flag1 := (Av^.z >= View.FustrumMinZ);
              Flag2 := (Bv^.z >= View.FustrumMinZ);
              Flag3 := (Cv^.z >= View.FustrumMinZ);

              if ((Flag1 And Flag2) And Flag3)
                Then Begin
                End
                Else Begin
                  ClipFlag := ClipFlag or CLIP_Z;
                  if ((Not(Flag1) And Not(Flag2)) and Not(Flag3))
                    Then TotalClip := 1;
                End;
            End;

          if ((Obj.ClipMask And CLIP_LEFT) <> 0)
            Then Begin
              Flag1 := (Av^.x >= -Av^.z);
              Flag2 := (Bv^.x >= -Bv^.z);
              Flag3 := (Cv^.x >= -Cv^.z);

              if ((Flag1 And Flag2) And Flag3)
                Then Begin
                End
                Else Begin
                  ClipFlag := ClipFlag or CLIP_LEFT;
                  if ((Not(Flag1) And Not(Flag2)) and Not(Flag3))
                    Then TotalClip := 1;
                End;
            End;

          if ((Obj.ClipMask And CLIP_RIGHT) <> 0)
            Then Begin
              Flag1 := (Av^.x <= Av^.z);
              Flag2 := (Bv^.x <= Bv^.z);
              Flag3 := (Cv^.x <= Cv^.z);

              if ((Flag1 And Flag2) And Flag3)
                Then Begin
                End
                Else Begin
                  ClipFlag := ClipFlag or CLIP_RIGHT;
                  if ((Not(Flag1) And Not(Flag2)) and Not(Flag3))
                    Then TotalClip := 1;
                End;
            End;

          if ((Obj.ClipMask And CLIP_BOTTOM) <> 0)
            Then Begin
              Flag1 := (Av^.y >= -Av^.z);
              Flag2 := (Bv^.y >= -Bv^.z);
              Flag3 := (Cv^.y >= -Cv^.z);

              if ((Flag1 And Flag2) And Flag3)
                Then Begin
                End
                Else Begin
                  ClipFlag := ClipFlag or CLIP_BOTTOM;
                  if ((Not(Flag1) And Not(Flag2)) and Not(Flag3))
                    Then TotalClip := 1;
                End;
            End;

          if ((Obj.ClipMask And CLIP_TOP) <> 0)
            Then Begin
              Flag1 := (Av^.y <= Av^.z);
              Flag2 := (Bv^.y <= Bv^.z);
              Flag3 := (Cv^.y <= Cv^.z);

              if ((Flag1 And Flag2) And Flag3)
                Then Begin
                End
                Else Begin
                  ClipFlag := ClipFlag or CLIP_TOP;
                  if ((Not(Flag1) And Not(Flag2)) and Not(Flag3))
                    Then TotalClip := 1;
                End;
            End;

          if (TotalClip = 0)
            Then Begin
              { polygon does need clipping, so lets do it }
              if ((ClipFlag And Not(CLIP_SHADEMASK)) > 0)
                Then FustrumClip(View, P, Obj.RenderMask, ClipFlag);

            End
            Else Begin
              { the polygon is completely outside the view fustrum }
              P^.Visible := 0;
            End;

        End;
    End;
End;

Procedure pCVert3dSwap(Var A, B : pCVert3d);

Var
  t : pCVert3d;

Begin
  t := a;
  a := b;
  b := t;
End;

Procedure pCVert2dSwap(Var A, B : pCVert2d);

Var
  t : pCVert2d;

Begin
  t := a;
  a := b;
  b := t;
End;

Procedure FustrumClip(View : ViewObject; P : pTriangle; ShadeIndex, Mask : Longint);


Var
  i, NumVert, ClipFlag : Longint;
  TempPoly  : pTriangle;
  InUV      : pCVert2d;
  OutUV     : pCVert2d;
  InVerts   : pCVert3d;
  OutVerts  : pCVert3d;

Begin

  For i := 0 to 15 do
    Begin
      IVerts[i] := NIL;
      OVerts[i] := NIL;
    End;

  For i := 0 to 2 do
    IVerts[i] := GrabVertexFromPool;

  IVerts[0]^.Cam := P^.A^.Cam;
  IVerts[1]^.Cam := P^.B^.Cam;
  IVerts[2]^.Cam := P^.C^.Cam;

  if ((Mask And CLIP_GOURAUD) > 0)
    Then Begin
      IVerts[0]^.Shade := P^.A^.Shade;
      IVerts[1]^.Shade := P^.B^.Shade;
      IVerts[2]^.Shade := P^.C^.Shade;
    End;

  if ((Mask And CLIP_ENV) > 0)
    Then Begin
      IVerts[0]^.Env := P^.A^.Env;
      IVerts[1]^.Env := P^.B^.Env;
      IVerts[2]^.Env := P^.C^.Env;
    End;

  if ((Mask And CLIP_TEXTURE) > 0)
    Then Begin
      InUV  := @IUV;
      OutUV := @OUV;
      InUV^[0] := P^.UV1;
      InUV^[1] := P^.UV2;
      InUV^[2] := P^.UV3;
    End;

  InVerts   := @IVerts;
  OutVerts  := @OVerts;

  NumVert   := 3;

  if ((Mask And CLIP_Z) > 0)
    Then Begin
      ClipFlag := 1;
      zClip(InVerts, OutVerts, InUV, OutUV, View, NumVert, Mask);
      pCVert3dSwap(InVerts, OutVerts);
      pCVert2dSwap(InUV, OutUV);
    End;

  if ((Mask And CLIP_LEFT) > 0) And (NumVert >= 3)
    Then Begin
      ClipFlag := 1;
      LeftClip(InVerts, OutVerts, InUV, OutUV, View, NumVert, Mask);
      pCVert3dSwap(InVerts, OutVerts);
      pCVert2dSwap(InUV, OutUV);
    End;

  if ((Mask And CLIP_RIGHT) > 0) And (NumVert >= 3)
    Then Begin
      ClipFlag := 1;
      RightClip(InVerts, OutVerts, InUV, OutUV, View, NumVert, Mask);
      pCVert3dSwap(InVerts, OutVerts);
      pCVert2dSwap(InUV, OutUV);
    End;

  if ((Mask And CLIP_BOTTOM) > 0) And (NumVert >= 3)
    Then Begin
      ClipFlag := 1;
      BottomClip(InVerts, OutVerts, InUV, OutUV, View, NumVert, Mask);
      pCVert3dSwap(InVerts, OutVerts);
      pCVert2dSwap(InUV, OutUV);
    End;

  if ((Mask And CLIP_TOP) > 0) And (NumVert >= 3)
    Then Begin
      ClipFlag := 1;
      TopClip(InVerts, OutVerts, InUV, OutUV, View, NumVert, Mask);
      pCVert3dSwap(InVerts, OutVerts);
      pCVert2dSwap(InUV, OutUV);
    End;

  if (ClipFlag > 0)
    Then P^.Visible := 0;

  if (ClipFlag > 0) And (NumVert >= 3)
    Then Begin
      for i := 0 to NumVert - 1 do
        SingleProject(View, InVerts^[i]);

      for i := 0 to NumVert - 3 do
        Begin
          TempPoly := GrabPolygonFromPool;

          TempPoly^.Color := P^.Color;
          TempPoly^.Shade := P^.Shade;
          TempPoly^.A     := InVerts^[0];
          TempPoly^.B     := InVerts^[i+1];
          TempPoly^.C     := InVerts^[i+2];

          if ((Mask And CLIP_ENV) > 0)
            Then Begin
              TempPoly^.Texture := P^.Texture;
              TempPoly^.IBuffer := P^.IBuffer;
            End;

          if ((Mask And CLIP_TEXTURE) > 0)
            Then Begin
              TempPoly^.Texture := P^.Texture;
              TempPoly^.UV1 := InUV^[0];
              TempPoly^.UV2 := InUV^[i+1];
              TempPoly^.UV3 := InUV^[i+2];
            End;

          AddPolygonToRenderList(TempPoly, ShadeIndex);
        End;
    End;
End;

Procedure zClip     (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint);

Var
  OutVertCount : Longint;
  Curr, Next   : Longint;
  CurrFlag     : Boolean;
  NextFlag     : Boolean;
  VertCount    : Longint;
  MinZ         : Single;
  tParm        : Single;

Begin
  OutVertCount := 0;
  VertCount    := NumVert;
  MinZ         := View.FustrumMinZ;

  CurrFlag := (InVerts^[0]^.Cam.z >= MinZ);

  For Curr := 0 to VertCount - 1 do
    Begin
      Next     := (Curr + 1) Mod VertCount;
      NextFlag := (InVerts^[Next]^.Cam.z >= MinZ);

      if (CurrFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          OutVerts^[OutVertCount]^.Cam   := InVerts^[Curr]^.Cam;
          OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env := InVerts^[Curr]^.Env;
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount] := InUV^[Curr];
            End;

          Inc(OutVertCount);
        End;


      if (CurrFlag <> NextFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          tParm := (MinZ - InVerts^[Curr]^.Cam.z) / (InVerts^[Next]^.Cam.z - InVerts^[Curr]^.Cam.z);

          OutVerts^[OutVertCount]^.Cam.x := InVerts^[Curr]^.Cam.x + tParm * (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          OutVerts^[OutVertCount]^.Cam.y := InVerts^[Curr]^.Cam.y + tParm * (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          OutVerts^[OutVertCount]^.Cam.z := MinZ;

          if ((ClipParm And CLIP_GOURAUD) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade + tParm * (InVerts^[Next]^.Shade-InVerts^[Curr]^.Shade);
            End;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env.x := InVerts^[Curr]^.Env.x + tParm * (InVerts^[Next]^.Env.x-InVerts^[Curr]^.Env.x);
              OutVerts^[OutVertCount]^.Env.y := InVerts^[Curr]^.Env.y + tParm * (InVerts^[Next]^.Env.y-InVerts^[Curr]^.Env.y);
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount].x := InUV^[Curr].x + tParm * (InUV^[Next].x - InUV^[Curr].x);
              OutUV^[OutVertCount].y := InUV^[Curr].y + tParm * (InUV^[Next].y - InUV^[Curr].y);
            End;

          Inc(OutVertCount);
        End;

      CurrFlag := NextFlag;
    End;

  NumVert := OutVertCount;
End;

Procedure LeftClip     (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint);

Var
  OutVertCount : Longint;
  Curr, Next   : Longint;
  CurrFlag     : Boolean;
  NextFlag     : Boolean;
  VertCount    : Longint;
  MinZ         : Single;
  tParm        : Single;
  dx, dy, dz   : Single;

Begin
  OutVertCount := 0;
  VertCount    := NumVert;
  MinZ         := View.FustrumMinZ;

  CurrFlag := (InVerts^[0]^.Cam.x >= (-InVerts^[0]^.Cam.z));

  For Curr := 0 to VertCount - 1 do
    Begin
      Next     := (Curr + 1) Mod VertCount;
      NextFlag := (InVerts^[Next]^.Cam.x >= (-InVerts^[Next]^.Cam.z));

      if (CurrFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          OutVerts^[OutVertCount]^.Cam := InVerts^[Curr]^.Cam;
          OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env := InVerts^[Curr]^.Env;
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount] := InUV^[Curr];
            End;

          Inc(OutVertCount);
        End;


      if (CurrFlag <> NextFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          dx := (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          dz := (InVerts^[Next]^.Cam.z - InVerts^[Curr]^.Cam.z);

          tParm := -(InVerts^[Curr]^.Cam.x + InVerts^[Curr]^.Cam.z) / (dx + dz);

          OutVerts^[OutVertCount]^.Cam.x := InVerts^[Curr]^.Cam.x + tParm * (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          OutVerts^[OutVertCount]^.Cam.y := InVerts^[Curr]^.Cam.y + tParm * (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          OutVerts^[OutVertCount]^.Cam.z := -OutVerts^[OutVertCount]^.Cam.x;

          if ((ClipParm And CLIP_GOURAUD) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade + tParm * (InVerts^[Next]^.Shade-InVerts^[Curr]^.Shade);
            End;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env.x := InVerts^[Curr]^.Env.x + tParm * (InVerts^[Next]^.Env.x-InVerts^[Curr]^.Env.x);
              OutVerts^[OutVertCount]^.Env.y := InVerts^[Curr]^.Env.y + tParm * (InVerts^[Next]^.Env.y-InVerts^[Curr]^.Env.y);
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount].x := InUV^[Curr].x + tParm * (InUV^[Next].x - InUV^[Curr].x);
              OutUV^[OutVertCount].y := InUV^[Curr].y + tParm * (InUV^[Next].y - InUV^[Curr].y);
            End;

          Inc(OutVertCount);
        End;

      CurrFlag := NextFlag;
    End;

  NumVert := OutVertCount;
End;


Procedure RightClip     (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint);

Var
  OutVertCount : Longint;
  Curr, Next   : Longint;
  CurrFlag     : Boolean;
  NextFlag     : Boolean;
  VertCount    : Longint;
  MinZ         : Single;
  tParm        : Single;
  dx, dy, dz   : Single;

Begin
  OutVertCount := 0;
  VertCount    := NumVert;
  MinZ         := View.FustrumMinZ;

  CurrFlag := (InVerts^[0]^.Cam.x <= InVerts^[0]^.Cam.z);

  For Curr := 0 to VertCount - 1 do
    Begin
      Next     := (Curr + 1) Mod VertCount;
      NextFlag := (InVerts^[Next]^.Cam.x <= InVerts^[Next]^.Cam.z);

      if (CurrFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          OutVerts^[OutVertCount]^.Cam   := InVerts^[Curr]^.Cam;
          OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env := InVerts^[Curr]^.Env;
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount] := InUV^[Curr];
            End;

          Inc(OutVertCount);
        End;


      if (CurrFlag <> NextFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          dx := (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          dz := (InVerts^[Next]^.Cam.z - InVerts^[Curr]^.Cam.z);

          tParm := (-InVerts^[Curr]^.Cam.x + InVerts^[Curr]^.Cam.z) / (dx - dz);

          OutVerts^[OutVertCount]^.Cam.x := InVerts^[Curr]^.Cam.x + tParm * (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          OutVerts^[OutVertCount]^.Cam.y := InVerts^[Curr]^.Cam.y + tParm * (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          OutVerts^[OutVertCount]^.Cam.z := OutVerts^[OutVertCount]^.Cam.x;

          if ((ClipParm And CLIP_GOURAUD) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade + tParm * (InVerts^[Next]^.Shade-InVerts^[Curr]^.Shade);
            End;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env.x := InVerts^[Curr]^.Env.x + tParm * (InVerts^[Next]^.Env.x-InVerts^[Curr]^.Env.x);
              OutVerts^[OutVertCount]^.Env.y := InVerts^[Curr]^.Env.y + tParm * (InVerts^[Next]^.Env.y-InVerts^[Curr]^.Env.y);
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount].x := InUV^[Curr].x + tParm * (InUV^[Next].x - InUV^[Curr].x);
              OutUV^[OutVertCount].y := InUV^[Curr].y + tParm * (InUV^[Next].y - InUV^[Curr].y);
            End;

          Inc(OutVertCount);
        End;

      CurrFlag := NextFlag;
    End;

  NumVert := OutVertCount;
End;


Procedure BottomClip     (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint);

Var
  OutVertCount : Longint;
  Curr, Next   : Longint;
  CurrFlag     : Boolean;
  NextFlag     : Boolean;
  VertCount    : Longint;
  MinZ         : Single;
  tParm        : Single;
  dx, dy, dz   : Single;

Begin
  OutVertCount := 0;
  VertCount    := NumVert;
  MinZ         := View.FustrumMinZ;

  CurrFlag := (InVerts^[0]^.Cam.y >= (-InVerts^[0]^.Cam.z));

  For Curr := 0 to VertCount - 1 do
    Begin
      Next     := (Curr + 1) Mod VertCount;
      NextFlag := (InVerts^[Next]^.Cam.y >= (-InVerts^[Next]^.Cam.z));

      if (CurrFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          OutVerts^[OutVertCount]^.Cam   := InVerts^[Curr]^.Cam;
          OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env := InVerts^[Curr]^.Env;
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount] := InUV^[Curr];
            End;

          Inc(OutVertCount);
        End;


      if (CurrFlag <> NextFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          dy := (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          dz := (InVerts^[Next]^.Cam.z - InVerts^[Curr]^.Cam.z);

          tParm := -(InVerts^[Curr]^.Cam.y + InVerts^[Curr]^.Cam.z) / (dy + dz);

          OutVerts^[OutVertCount]^.Cam.x := InVerts^[Curr]^.Cam.x + tParm * (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          OutVerts^[OutVertCount]^.Cam.y := InVerts^[Curr]^.Cam.y + tParm * (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          OutVerts^[OutVertCount]^.Cam.z := -OutVerts^[OutVertCount]^.Cam.y;

          if ((ClipParm And CLIP_GOURAUD) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade + tParm * (InVerts^[Next]^.Shade-InVerts^[Curr]^.Shade);
            End;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env.x := InVerts^[Curr]^.Env.x + tParm * (InVerts^[Next]^.Env.x-InVerts^[Curr]^.Env.x);
              OutVerts^[OutVertCount]^.Env.y := InVerts^[Curr]^.Env.y + tParm * (InVerts^[Next]^.Env.y-InVerts^[Curr]^.Env.y);
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount].x := InUV^[Curr].x + tParm * (InUV^[Next].x - InUV^[Curr].x);
              OutUV^[OutVertCount].y := InUV^[Curr].y + tParm * (InUV^[Next].y - InUV^[Curr].y);
            End;

          Inc(OutVertCount);
        End;

      CurrFlag := NextFlag;
    End;

  NumVert := OutVertCount;
End;


Procedure TopClip     (InVerts, OutVerts : pCVert3d; InUV, OutUV : pCVert2d;
                         View : ViewObject; Var NumVert : Longint; ClipParm : Longint);

Var
  OutVertCount : Longint;
  Curr, Next   : Longint;
  CurrFlag     : Boolean;
  NextFlag     : Boolean;
  VertCount    : Longint;
  MinZ         : Single;
  tParm        : Single;
  dx, dy, dz   : Single;

Begin
  OutVertCount := 0;
  VertCount    := NumVert;
  MinZ         := View.FustrumMinZ;

  CurrFlag := (InVerts^[0]^.Cam.y <= InVerts^[0]^.Cam.z);

  For Curr := 0 to VertCount - 1 do
    Begin
      Next     := (Curr + 1) Mod VertCount;
      NextFlag := (InVerts^[Next]^.Cam.y <= InVerts^[Next]^.Cam.z);

      if (CurrFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          OutVerts^[OutVertCount]^.Cam   := InVerts^[Curr]^.Cam;
          OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env := InVerts^[Curr]^.Env;
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount] := InUV^[Curr];
            End;

          Inc(OutVertCount);
        End;


      if (CurrFlag <> NextFlag)
        Then Begin
          if (OutVerts^[OutVertCount] = NIL)
            Then OutVerts^[OutVertCount] := GrabVertexFromPool;

          dy := (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          dz := (InVerts^[Next]^.Cam.z - InVerts^[Curr]^.Cam.z);

          tParm := (-InVerts^[Curr]^.Cam.y + InVerts^[Curr]^.Cam.z) / (dy - dz);

          OutVerts^[OutVertCount]^.Cam.x := InVerts^[Curr]^.Cam.x + tParm * (InVerts^[Next]^.Cam.x - InVerts^[Curr]^.Cam.x);
          OutVerts^[OutVertCount]^.Cam.y := InVerts^[Curr]^.Cam.y + tParm * (InVerts^[Next]^.Cam.y - InVerts^[Curr]^.Cam.y);
          OutVerts^[OutVertCount]^.Cam.z := OutVerts^[OutVertCount]^.Cam.y;

          if ((ClipParm And CLIP_GOURAUD) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Shade := InVerts^[Curr]^.Shade + tParm * (InVerts^[Next]^.Shade-InVerts^[Curr]^.Shade);
            End;

          if ((ClipParm And CLIP_ENV) > 0)
            Then Begin
              OutVerts^[OutVertCount]^.Env.x := InVerts^[Curr]^.Env.x + tParm * (InVerts^[Next]^.Env.x-InVerts^[Curr]^.Env.x);
              OutVerts^[OutVertCount]^.Env.y := InVerts^[Curr]^.Env.y + tParm * (InVerts^[Next]^.Env.y-InVerts^[Curr]^.Env.y);
            End;

          if ((ClipParm And CLIP_TEXTURE) > 0)
            Then Begin
              OutUV^[OutVertCount].x := InUV^[Curr].x + tParm * (InUV^[Next].x - InUV^[Curr].x);
              OutUV^[OutVertCount].y := InUV^[Curr].y + tParm * (InUV^[Next].y - InUV^[Curr].y);
            End;

          Inc(OutVertCount);
        End;

      CurrFlag := NextFlag;
    End;

  NumVert := OutVertCount;
End;

End.