Unit Memory;
{
  A C-Like memory manager.

  Supplements the Pascal memory manager with a much nicer interface
  which removes the ugly parameters associated with Getmem() and Fremem()

  Alex Chalfin
  6/1/97
}

Interface

Function  Malloc(Size : Longint) : Pointer;
Procedure Free(P : Pointer);
Function  Realloc(P : Pointer; Size : Longint) : Pointer;

Implementation

Type
  pMemoryRec = ^MemoryRec;
  MemoryRec = Record
    Size : Longint;
    Loc  : Pointer;
    Next : pMemoryRec;
  End;

Const
  MemList     : pMemoryRec = NIL;

{===========================================================================}
{ Malloc                                                                    }
{===========================================================================}
Function  Malloc(Size : Longint) : Pointer;

Var
  TempMemRec    : pMemoryRec;
  TempPtr       : Pointer;

Begin
  { Check and see if enough memory exists }
  if (MaxAvail < Size)
    Then TempPtr := NIL
    Else Getmem(TempPtr, Size);

  { if TempPtr = NIL, memory could not be allocated so return NIL pointer }
  if (TempPtr = NIL)
    Then Begin
      Malloc := NIL;
      Exit;
    End;

  { Allocate a new memory record for recording memory }
  New(TempMemRec);

  { if the memory record couldn't be allocated, return NIL }
  if (TempMemRec = NIL)
    Then Begin
      Freemem(TempPtr, Size);
      Malloc := NIL;
      Exit;
    End;

  { Assign the values to the structure }
  TempMemRec^.Size := Size;
  TempMemRec^.Loc  := TempPtr;
  TempMemRec^.Next := MemList;

  MemList := TempMemRec;

  Malloc := TempPtr;
End;

{===========================================================================}
{ Free                                                                      }
{===========================================================================}
Procedure Free(P : Pointer);

Var
  Traverse : pMemoryRec;
  Prev     : pMemoryRec;

Begin
  Traverse := MemList;
  Prev     := MemList;

  { Traverse the memory list to find the allocated pointer }
  While (Traverse <> NIL) do
    Begin
      if (Traverse^.Loc = P)
        Then Begin
          Freemem(Traverse^.Loc, Traverse^.Size);

          if (Prev = Traverse)
            Then Begin
              MemList := Traverse^.Next;
              Dispose(Traverse);
            End
            Else Begin
              Prev^.Next := Traverse^.Next;
              Dispose(Traverse);
            End;

          Exit;
        End;

      Prev     := Traverse;
      Traverse := Traverse^.Next;
    End;
End;

{===========================================================================}
{ Realloc                                                                   }
{===========================================================================}
Function  Realloc(P : Pointer; Size : Longint) : Pointer;

Var
  Traverse : pMemoryRec;
  TempPtr : Pointer;

Begin
  Traverse := MemList;

  While (Traverse <> NIL) do
    Begin
      if (Traverse^.Loc = P)
        Then Begin
          TempPtr := Malloc(Size); { allocate new area }

          if (TempPtr = NIL)    { check to see if it allocated }
            Then Begin
              Realloc := NIL;
              Exit;
            End;

          if (Traverse^.Size < Size)
            Then Move(Traverse^.Loc^, TempPtr^, Traverse^.Size)
            Else Move(Traverse^.Loc^, TempPtr^, Size);   { copy the data over }

          Freemem(Traverse^.Loc, Traverse^.Size); { free the old data }
          Traverse^.Loc  := TempPtr;     { assign to new size and location }
          Traverse^.Size := Size;

          Realloc := TempPtr;
          Exit;
        End;

      Traverse := Traverse^.Next;
    End;
  Realloc := NIL;
End;

End.