(*

                     Inertia Realtime 3D Rendering Engine
     Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*)


Unit PCX;
{$I-}

Interface

Uses
  Types;

Function LoadPCXFile(Filename : String; Var BitMap : Pointer; Var Pal : Palette; Var Size : Longint; FOfs : Longint) : Integer;
Procedure SavePCXFile(Filename : String; XSize, YSize : Word; Screen : Pointer; Pal : Palette);

Implementation

Uses
  CFile,
  Memory;

Type
  PCXHeaderType = Record
    Manufacturer : Byte;
    Version      : Byte;
    Encoding     : Byte;
    BitsPerPixel : Byte;
    XMin, YMin, XMax, YMax : Integer;
    HDpi, VDpi   : Integer;
    ColorMap     : Array[0..47] of Byte;
    Reserved     : Byte;
    NPlanes      : Byte;
    BytesPerLine : Integer;
    PaletteInfo  : Integer;
    HScreenSize  : Integer;
    VScreenSize  : Integer;
    Filler       : Array[0..53] of Byte;
  End;


Procedure SavePCXFile(Filename : String; XSize, YSize : Word; Screen : Pointer; Pal : Palette);

Begin
End;

Function LoadPCXFile(Filename : String; Var BitMap : Pointer; Var Pal : Palette; Var Size : Longint; FOfs : Longint) : Integer;
{ Returns:  0 - Operation successful    }
{           1 - File not found          }
{           2 - Error reading from file }
{           3 - Not enough memory       }
{           4 - Not a 256 color file    }

Var
  Header : PCXHeaderType;
  TempPtr : Pointer;
  pSize : Longint;
  TextureOffset : Longint;
  RunLen : Byte;
  Value  : Byte;
  InRead : Longint;
  Buffer : Array[0..255] of Byte;
  TempAddr : DWord;
  fp : ^File;

Begin
  fp := FOpen(FileName, 'r');

  if (fp = NIL)
    Then Begin
      LoadPCXFile := 1;
      Exit;
    End;

  FSeek(fp, FOfs, SEEK_SET);
  InRead := FRead(Header, 1, Sizeof(Header), fp);
  If IOResult <> 0
    Then Begin
      FClose(fp);
      LoadPCXFile := 2;
      Exit;
    End;
  If (Header.BitsPerPixel <> 8)
    Then Begin
      FClose(fp);
      LoadPCXFile := 4;
      Exit;
    End;
  pSize := (Header.XMax-Header.XMin+1)*(Header.yMax-Header.YMin+1);

  TempPtr := Malloc(pSize);
  TempAddr := DWord(TempPtr);

  TextureOffset := 0;
  While TextureOffset < pSize do
    Begin
      InRead := FRead(RunLen, 1, sizeof(RunLen), fp);
      If IOResult <> 0
        Then Begin
          FClose(fp);
          Free(TempPtr);
          LoadPCXFile := 2;
          Exit;
        End;
      If (RunLen and $C0) = $C0
        Then Begin
          RunLen := RunLen And $3f;
          InRead := FRead(Value, 1, sizeof(Value), fp);
          If IOResult <> 0
            Then Begin
              FClose(fp);
              Free(TempPtr);
              LoadPCXFile := 2;
              Exit;
            End;
        End
        Else Begin
          Value := RunLen;
          RunLen := 1;
        End;
      While (RunLen >= 1) and (TextureOffset < pSize) do
        Begin
          Mem[TempAddr] := Value;
          Inc(TempAddr);
          TextureOffset := TextureOffset + 1;
          RunLen := RunLen - 1;
        End;
    End;
  Repeat
    InRead := FRead(Value, 1, sizeof(Value), fp);
    If IOResult <> 0
      Then Begin
        FClose(fp);
        Free(TempPtr);
        LoadPCXFile := 2;
        Exit;
      End;
  Until (Value = 12);
  InRead := FRead(Pal, 1, Sizeof(Palette), fp);
  If IOResult <> 0
    Then Begin
      FClose(fp);
      Free(TempPtr);
      LoadPCXFile := 2;
      Exit;
    End;
  FClose(fp);

  For Value := 0 to 255 do
    Begin
      Pal[Value].r := Pal[Value].r Div 4;
      Pal[Value].g := Pal[Value].g Div 4;
      Pal[Value].b := Pal[Value].b Div 4;
    End;
  Size := pSize;
  BitMap := TempPtr;
  LoadPCXFile := 0;
End;

End.