;
;                    Inertia Realtime 3D Rendering Engine
;    Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;

;
; XFORM.ASM
; code: Alex Chalfin
;

.486p                           ; 486 protected mode
.Model Flat,Pascal              ; with flat memory model
                                ;
                                ;
DATA SEGMENT PARA PUBLIC USE32  ;
                                ;
include strucdef.asm            ;  Inertia structure references
                                ;
Public Transform                ;
Public Project                  ;
                                ;
extrn  AspectRatio : DWord      ;
                                ;
_tmp_flt      dd ?              ;
_tmp_flt_half dd 0.5            ;
                                ;
ENDS                            ;
                                ;
;================================================================
; Code Segment
;================================================================
                                ;
CODE SEGMENT PARA PUBLIC USE32  ;
                                ;
;===========================================================================
; Macros
;===========================================================================
                                ;
                                ;
XFormComponent Macro M0,M1,M2,M3,VLOC
  fld  st(2)                    ; re-load the x-coord
  fmul dword ptr [edi+M0*4]     ; multiply by Matrix[M0]
  fld  st(2)                    ; re-load the y-coord
  fmul dword ptr [edi+M1*4]     ; Multiply by Matrix[M1]
  fld  st(2)                    ; re-load the Z-coord
  fmul dword ptr [edi+M2*4]     ; Multiply by Matrix[M2]
                                ;
  fadd dword ptr [edi+M3*4]     ; Add in Matrix[M3]
  faddp st(1),st(0)             ;
  faddp st(1),st(0)             ;
  fstp  dword ptr VLOC          ; Store the camera Coordinate
EndM                            ;
                                ;
                                ;
;===========================================================================
; Transform
;===========================================================================
Transform Proc Uses edi esi ebx,Object : DWord, Matrix : DWord

  Mov  edi,DWord Ptr [Object]   ;
  Mov  ecx,[edi].NumVert        ;  Grab the number of verticies in object
                                ;
  Cmp  ecx,0                    ;  Check to see if there are verticies
  Jle @@ExitNoXFormVerticies    ;  to transform
                                ;
  Mov  esi,[edi].Vertexs        ;  Grab the pointer to the verticies
                                ;
  Mov  edi,DWord Ptr [Matrix]   ;  get the pointer to the matrix
                                ;
 @@XFormLooper:                 ;
  Mov  eax,[esi].VVisible       ;  grab the visible flag
                                ;
  Cmp  eax,1                    ;  Check to see if the vertex is indeed
  Jne  @@DontXFormVertex        ;  visible
                                ;
  fld  dword ptr [esi].Locx     ;  Save the X,Y,Z locations on the FPU
  fld  dword ptr [esi].Locy     ;  stack for faster access
  fld  dword ptr [esi].Locz     ;
                                ;
  XFormComponent 0,3,6,9, [esi].Camx  ; Calculate the Camera X value
  XFormComponent 1,4,7,10,[esi].Camy  ; Calculate the Camera Y value
  XFormComponent 2,5,8,11,[esi].Camz  ; Calculate the Camera Z value
                                ;
  fstp  dword ptr [_tmp_flt]    ; pop the z-coord
  fstp  dword ptr [_tmp_flt]    ; pop the y-coord
  fstp  dword ptr [_tmp_flt]    ; pop the x-coord
                                ;
                                ; FPU stack is now empty
 @@DontXFormVertex:             ;
  Add  esi,VERTEX_SIZE          ; go to the next vertex
  Dec  ecx                      ;
  Jnz  @@XFormLooper            ;
                                ;
 @@ExitNoXFormVerticies:        ;
  Ret                           ;
Transform EndP                  ;

;===========================================================================
; Project
;===========================================================================
Project Proc Uses edi esi ebx, View : DWord, Object : DWord

  Mov  edi,DWord Ptr [View]     ;
  Mov  esi,DWord Ptr [Object]   ;
                                ;
  Mov  ecx,[esi].NumVert        ;
  Cmp  ecx,0                    ;
  Jle @@ExitNoProjectionVert    ;
                                ;
  fld  dword ptr [edi].CenterX  ;
  fld  dword ptr [edi].CenterY  ; Store the center X and Y on the FPU stack
                                ;
  Mov  eax,[edi].WinMaxX        ; Set up the scaling factor for projectiona
  Sub  eax,[edi].WinMinX        ;
  Inc  eax                      ;
                                ;
  Mov  [_tmp_flt],eax           ;
  fild dword ptr [_tmp_flt]     ;
  fmul dword ptr [_tmp_flt_half];  calculate XScale, keep on stack
                                ;
  fld  st(0)                    ;  re-load it
                                ;
  fmul dword ptr [edi].InvCamYFac;
  fmul dword ptr [AspectRatio]  ;  YScale, keep on stack
                                ;
                                ; FPU Stack:
                                ;   0    |   1    |    2    |    3
                                ; YScale | XScale | CenterY | CenterX
                                ;
  Mov  edi,[esi].Vertexs        ; Get the vertex pointer
 @@ProjectLooper:               ;
                                ;
  Mov  eax,[edi].VVisible       ; check to see if the vertex is visible
  Cmp  eax,1                    ;
  Jne @@SkipProjection          ; if not, skip its projection
                                ;
  fld1                          ;
  fdiv  dword ptr [edi].Camz    ; calculate 1/z term
                                ;
  fst   dword ptr [edi].OneOverZ; store one over z  (5 things on the stack)
                                ;
  fld   st(2)                   ; re-load XScale    (6)
  fmul  dword ptr [edi].Camx    ; multiply by the coordinate X-component
  fmul  st(0),st(1)             ; multiply by the OneOverZ value
  fadd  st(0),st(5)             ; Add in CenterX
                                ;
  fstp  dword ptr [edi].ScrX    ; store x-component (5)
                                ;
  fld   st(1)                   ; re-load YScale    (6)
  fmul  dword ptr [edi].Camy    ; multiply by the coordinate X-component
  fmul  st(0),st(1)             ; multiply by the OneOverZ value
  fld   st(4)                   ; re-load CenterY   (7)
  fsubrp st(1),st(0)            ;                   (6)
                                ;
  fstp  dword ptr [edi].ScrY    ; store y-component (5)
                                ;
  fstp  dword ptr [_tmp_flt]    ; get the old 1/z value off (4)
                                ;
 @@SkipProjection:              ;
                                ;
  Add  edi,VERTEX_SIZE          ;
  Dec  ecx                      ;
  Jnz @@ProjectLooper           ;
                                ;  Clear the values of the FPU stack
  fstp dword ptr [_tmp_flt]     ;  (3)
  fstp dword ptr [_tmp_flt]     ;  (2)
  fstp dword ptr [_tmp_flt]     ;  (1)
  fstp dword ptr [_tmp_flt]     ;  (0)
                                ;
 @@ExitNoProjectionVert:        ;
  Ret                           ;
Project EndP                    ;


ENDS
END

Comment 
============================================================================
  Pascal Versions of each of the functions
============================================================================

Procedure TransForm(Var Obj : VectorObject; M : Matrix4x3);

Var
  i : Longint;
  V : pVertex;

Begin
  For i := 0 to Obj.NumVert - 1 do
    Begin
      V := @Obj.Vertexs^[i];

      if (V^.Visible = 1)
        Then Begin
          V^.Cam.x := (V^.Loc.x * M[0]) + (V^.Loc.y * M[3]) + (V^.Loc.z * M[6]) + M[XLOC];
          V^.Cam.y := (V^.Loc.x * M[1]) + (V^.Loc.y * M[4]) + (V^.Loc.z * M[7]) + M[YLOC];
          V^.Cam.z := (V^.Loc.x * M[2]) + (V^.Loc.y * M[5]) + (V^.Loc.z * M[8]) + M[ZLOC];
        End;
    End;
End;

Procedure Project(Var View : ViewObject; Var Obj : VectorObject);

Var
  i : Longint;
  V : pVertex;
  XScale : Single;
  YScale : Single;
  WinX   : Single;

Begin
  WinX := View.WinMaxX - View.WinMinX + 1;

  XScale  := (WinX / 2.0);
  YScale  := XScale * AspectRatio * View.InvCamYFac;

  For i := 0 to Obj.NumVert - 1 do
    Begin
      V := @Obj.Vertexs^[i];

      if (V^.Visible = 1)
        Then Begin
          V^.OneOverZ := 1.0 / V^.Cam.z;
          V^.Scr.x := (View.CenterX + (XScale * V^.Cam.x * V^.OneOverZ));
          V^.Scr.y := (View.CenterY - (YScale * V^.Cam.y * V^.OneOverZ));
        End;
    End;
End;



