(*

                       Driver8 Realtime 3D Rendering Engine
            Copyright (c) 1997, Alex Chalfin. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*)

Unit XForm;

Interface

Uses
  Types;

Procedure SingleProject(Var View : ViewObject; V : pVertex);
Procedure TransForm(Var Obj : VectorObject; M : Matrix4x3);
Procedure Project(Var View : ViewObject; Var Obj : VectorObject);
Procedure InvXForm(Var V : Vertex3d; M : Matrix4x3);
Procedure SingleRotate(Var Vector : Vertex3d; xa, ya, za : Single);

Implementation

{ External Assembly Functions }
{$L XFORM.OBJ}
Procedure TransForm(Var Obj : VectorObject; M : Matrix4x3); External;
Procedure Project(Var View : ViewObject; Var Obj : VectorObject); External;


Procedure SingleProject(Var View : ViewObject; V : pVertex);

Var
  YScale : Single;
  XScale : Single;
  WinX   : Single;

Begin
  WinX := View.WinMaxX - View.WinMinX + 1;

  XScale  := (WinX / 2.0);
  YScale  := XScale * AspectRatio * View.InvCamYFac;

  V^.OneOverZ := 1.0 / V^.Cam.z;
  V^.Scr.x := (View.CenterX + (XScale * V^.Cam.x * V^.OneOverZ));
  V^.Scr.y := (View.CenterY - (YScale * V^.Cam.y * V^.OneOverZ));
End;

Procedure InvXForm(Var V : Vertex3d; M : Matrix4x3);

Var
  tx, ty, tz : Single;

Begin
  tx := V.x;
  ty := V.y;
  tz := V.z;

  V.x := M[0] * tx + M[1] * ty + M[2] * tz;
  V.y := M[3] * tx + M[4] * ty + M[5] * tz;
  V.z := M[6] * tx + M[7] * ty + M[8] * tz;
End;

Procedure SingleRotate(Var Vector : Vertex3d; xa, ya, za : Single);

Var
  tx, ty, tz : Single;
  sx, sy, sz : Single;
  cx, cy, cz : Single;

Begin
  sx := sin(xa);   sy := sin(ya);   sz := sin(za);
  cx := cos(xa);   cy := cos(ya);   cz := cos(za);

  tx := (Vector.x * cz) - (Vector.y * sz);
  ty := (Vector.x * sz) + (Vector.y * cz);
  tz := Vector.z;

  Vector.x := (tx * cy) + (tz * sy);
  Vector.y := ty;
  Vector.z := (tz * cy) - (tx * sy);

  tx := Vector.x;
  ty := (Vector.y * cx) - (Vector.z * sx);
  tz := (Vector.y * sx) + (Vector.z * cx);

  Vector.x := tx;
  Vector.y := ty;
  Vector.z := tz;
End;

End.

